package cmd

import (
	"fmt"

	flags "github.com/orus-io/go-flags"

	redner "orus.io/orus-io/rednerd/lib"
)

var MailSender redner.MailSender

// Run ...
func Run() error {
	parser, err := SetupParser()
	if err != nil {
		return err
	}

	return RunParser(parser)
}

func RunParser(parser *flags.Parser) error {
	defer func() {
		for _, cb := range OnShutdownCB {
			cb()
		}
	}()

	if _, err := parser.Parse(); err != nil {
		if fe, ok := err.(*flags.Error); ok {
			if fe.Type == flags.ErrHelp {
				fmt.Println(err)
			} else {
				return fmt.Errorf("failed to parse: %w", err)
			}
		} else {
			log.Err(err).Msg("failure")

			return err
		}
	}

	return nil
}

func SetupParser() (*flags.Parser, error) {
	parser := flags.NewNamedParser("rednerd", flags.HelpFlag|flags.PassDoubleDash)
	parser.ShortDescription = "Rednerd API"
	parser.LongDescription = "The Rednerd rendering server"
	parser.EnvNamespace = "REDNERD"

	options := NewOptions(parser)
	if _, err := parser.AddGroup("Global", "", options); err != nil {
		return nil, err
	}

	engineOptionsGroup, err := parser.AddGroup("Engines", "Engine specific settings", &engineOptions)
	if err != nil {
		return nil, err
	}
	engineOptionsGroup.Namespace = "engines"
	engineOptionsGroup.EnvNamespace = "ENGINES"

	if err := engineOptions.SetupEngineOptions(engineOptionsGroup); err != nil {
		return nil, err
	}

	mailSenderGroup, err := parser.AddGroup("Mail", "Mail options", &MailSender)
	if err != nil {
		return nil, err
	}
	mailSenderGroup.Namespace = "mail"
	mailSenderGroup.EnvNamespace = "MAIL"

	newSentryOptions(options).AddToParser(parser)

	serveCmd := NewServerCmd(options)
	if err := serveCmd.AddToParser(parser); err != nil {
		return nil, err
	}

	migrateCmd := MigrateCmd{
		options: options,
	}
	if _, err := parser.AddCommand(
		"migrate", "Init/migrate the database schema", "", &migrateCmd,
	); err != nil {
		return nil, fmt.Errorf("could not init the cmd parser: %w", err)
	}
	adminPasswordCmd := AdminPasswordCmd{
		options: options,
	}
	if _, err := parser.AddCommand(
		"admin-password", "Set the admin password", "", &adminPasswordCmd,
	); err != nil {
		return nil, fmt.Errorf("could not init the cmd parser: %w", err)
	}

	configFile := NewConfigWrite(parser)

	if _, err := parser.AddCommand(
		"generate-config", "Generate a configuration file", "", configFile,
	); err != nil {
		return nil, fmt.Errorf("could not init the cmd parser: %w", err)
	}

	var varlistCmd VarlistCmd
	if _, err := parser.AddCommand(
		"varlist", "", "", &varlistCmd,
	); err != nil {
		return nil, fmt.Errorf("could not init the cmd parser: %w", err)
	}

	return parser, nil
}
