package mustache

import (
	"context"
	"reflect"

	"github.com/cbroglie/mustache"
	"github.com/rs/zerolog"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/rendering"
)

// Engine is a TemplateEngine implementation on top of github.com/cbroglie/mustache.
type Engine struct{}

// Info ...
func (e Engine) Info() rendering.TemplateEngineInfo {
	return rendering.TemplateEngineInfo{
		Type:        "mustache",
		Language:    "mustache",
		Description: "Render a [mustache](https://mustache.github.io/) template.",
	}
}

// Load loads a template.
func (e Engine) Load(ctx context.Context, template *models.Template) (rendering.Template, error) {
	s, err := template.String()
	if err != nil {
		return nil, err
	}
	bodyTemplate, err := mustache.ParseString(s)
	if err != nil {
		return nil, err
	}
	t := Template{
		outputType:        template.Produces,
		metadata:          template.StaticMetadata,
		metadataTemplates: make(map[string]*mustache.Template),
		bodyTemplate:      bodyTemplate,
	}
	for name, meta := range template.Metadata {
		tmpl, err := mustache.ParseString(meta)
		if err != nil {
			return nil, err
		}
		t.metadataTemplates[name] = tmpl
	}

	return &t, nil
}

// Template ...
type Template struct {
	outputType        string
	metadata          models.Metadata
	metadataTemplates map[string]*mustache.Template
	bodyTemplate      *mustache.Template
}

type AutoContext struct {
	children      map[string]*AutoContext
	childrenOrder []string
}

func NewAutoContext() *AutoContext {
	return &AutoContext{
		children: make(map[string]*AutoContext),
	}
}

func (c *AutoContext) Lookup(name string) (reflect.Value, error) {
	value, ok := c.children[name]
	if !ok {
		value = NewAutoContext()
		c.children[name] = value
		c.childrenOrder = append(c.childrenOrder, name)
	}

	return reflect.ValueOf(value), nil
}

func (c *AutoContext) List() []string {
	return c.list()
}

func (c *AutoContext) list() []string {
	var res []string
	for _, name := range c.childrenOrder {
		sublist := c.children[name].list()
		if len(sublist) != 0 {
			for _, varname := range sublist {
				res = append(res, name+"."+varname)
			}
		} else {
			res = append(res, name)
		}
	}

	return res
}

func (t Template) Varlist() ([]string, error) {
	context := NewAutoContext()
	_, err := t.bodyTemplate.Render(context)
	if err != nil {
		return nil, err
	}

	return context.list(), nil
}

// Render the template.
func (t Template) Render(
	ctx context.Context, w rendering.DocumentWriter, record models.Record,
) error {
	w.Metadata().Update(t.metadata)

	for key, tmpl := range t.metadataTemplates {
		s, err := tmpl.Render(record)
		if err != nil {
			return err
		}
		w.Metadata()[key] = s
	}

	if err := t.bodyTemplate.FRender(w, record); err != nil {
		return err
	}

	return nil
}

func init() {
	rendering.RegisterTemplateEngineSetup(
		"Mustache",
		func() interface{} {
			return nil
		},
		func(optionGroup interface{}, log zerolog.Logger) (rendering.TemplateEngine, error) {
			return &Engine{}, nil
		},
	)
}
