package redner

import (
	"context"
	"database/sql"
	"errors"
	"fmt"
	"sync"
	"time"

	"github.com/Masterminds/squirrel"
	"github.com/jmoiron/sqlx"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/rs/zerolog"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/rendering"
	"orus.io/orus-io/rednerd/utils"
)

const LogRenderBatchSize = 1000

var (
	ErrInvalidInput = errors.New("invalid input")

	renderPipelineDuration = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "rednerd_render_pipeline",
			Help:       "Render pipeline duration and count",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
		},
		[]string{
			"template_language",
			"from_type",
			"to_type",
			"status",
		},
	)
)

func NewRenderer(
	ctx context.Context,
	db *sqlx.DB,
	registry *rendering.EngineRegistry,
	templateRegistry *rendering.TemplateEngineRegistry,
	log zerolog.Logger,
) *Renderer {
	r := Renderer{
		db:               db,
		registry:         registry,
		templateRegistry: templateRegistry,
		logRender:        make(chan *models.RenderLog, LogRenderBatchSize*2),
	}
	ctx, r.stop = context.WithCancel(ctx)
	r.wg.Add(1)

	go r.run(ctx, db, log)

	return &r
}

type Renderer struct {
	db               *sqlx.DB
	registry         *rendering.EngineRegistry
	templateRegistry *rendering.TemplateEngineRegistry
	logRender        chan *models.RenderLog
	wg               sync.WaitGroup
	stop             func()
}

func (r *Renderer) run(ctx context.Context, dbc *sqlx.DB, log zerolog.Logger) {
	defer r.wg.Done()

	var batch []database.Mapped

	db := database.NewSQLHelper(ctx, dbc, log)

	timer := time.NewTimer(0)
	defer timer.Stop()

	flush := func() {
		if len(batch) != 0 {
			if _, err := db.InsertNoPKey(batch...); err != nil {
				log.Err(err).Msg("failed to insert render logs")
			} else {
				log.Debug().Msg("successfully inserted render logs")
			}
			batch = batch[:0]
		}
	}

	for {
		select {
		case rlog := <-r.logRender:
			if rlog != nil {
				batch = append(batch, rlog)
				log.Debug().Msg("batched a render log")
			}
			if rlog == nil || len(batch) == LogRenderBatchSize {
				flush()
			}
			if rlog == nil {
				return
			}
			if len(batch) != 0 {
				timer.Reset(50 * time.Millisecond)
			}
		case <-timer.C:
			flush()
		case <-ctx.Done():
			close(r.logRender)
			flush()

			return
		}
	}
}

func (r *Renderer) Close() {
	r.stop()
	r.wg.Wait()
}

func (r *Renderer) Render(
	ctx context.Context,
	username string,
	template *models.Template,
	data models.Dataset,
	document *models.Document,
	metadata models.Metadata,
	toType string,
) ([]*models.Document, error) {
	var (
		log    = zerolog.Ctx(ctx)
		result []*models.Document

		templatedb       models.TemplateDB
		templateLanguage string
		fromType         string
	)

	pipelineTimer := utils.NewVecTimer(renderPipelineDuration)

	db := database.NewSQLHelper(ctx, r.db, *log)

	if template != nil {
		if template.Account != "" && template.Name != "" {
			if err := db.GetWhere(
				&templatedb,
				squirrel.Eq{
					models.TemplateDBAccountColumn: template.Account,
					models.TemplateDBNameColumn:    template.Name,
				},
			); err != nil {
				if err == sql.ErrNoRows {
					return nil, fmt.Errorf(
						"%w: no such template '%s/%s'",
						ErrInvalidInput, template.Account, template.Name)
				}

				return nil, err
			}
			templatedb.FromDB()
			template = &templatedb.Template
		}
		templateLanguage = template.Language
		fromType = template.Produces

		pipeline, err := r.templateRegistry.NewPipeline(
			ctx, r.registry, template, metadata, toType,
		)
		if err != nil {
			if err == rendering.ErrUnknownTemplateLanguage {
				return nil, fmt.Errorf("%w: unknown template language: %s", ErrInvalidInput, template.Language)
			}

			return nil, fmt.Errorf("%w: error loading rendering pipeline: %s", ErrInvalidInput, err)
		}
		for _, record := range data {
			var doc models.Document
			w := rendering.NewDocumentWriter(&doc)
			if err := pipeline.Render(ctx, w, record); err != nil {
				pipelineTimer.ObserveDurationWithLabelValues(
					templateLanguage, fromType, toType, "failure")

				return nil, err
			}
			result = append(result, &doc)
		}
	}

	if document != nil {
		pipeline, err := r.registry.GetPipeline(document.Type, toType)
		if err != nil {
			return nil, fmt.Errorf("%w: %s", ErrInvalidInput, err)
		}

		document.UpdateMetadata(metadata)

		if len(pipeline) != 0 {
			var output models.Document

			reader := rendering.NewDocumentReader(document)
			writer := rendering.NewDocumentWriter(&output)

			if err := pipeline.Render(ctx, writer, reader); err != nil {
				return nil, err
			}
			result = append(result, &output)
		} else {
			result = append(result, document)
		}
	}

	d := pipelineTimer.ObserveDurationWithLabelValues(
		templateLanguage, fromType, toType, "success")

	r.logRender <- models.NewRenderLog(
		pipelineTimer.GetBegin(), d,
		fromType, toType, username,
		templatedb.ID, templatedb.InternalVersion, templateLanguage,
	)

	return result, nil
}

func init() {
	prometheus.MustRegister(renderPipelineDuration)
}
