package handlers

import (
	"context"
	"errors"
	"fmt"
	"strings"
	"time"

	"github.com/go-openapi/runtime/middleware"
	"github.com/go-openapi/strfmt"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/restapi/operations/template_management"
)

// NewAccountTemplateAddHandler returns a AccountTemplateAddHandler.
func NewAccountTemplateAddHandler(db *sqlx.DB) *AccountTemplateAddHandler {
	return &AccountTemplateAddHandler{
		db: db,
	}
}

// AccountTemplateAddHandler handles GET /template.
type AccountTemplateAddHandler struct {
	db *sqlx.DB
}

func (h *AccountTemplateAddHandler) handle(
	ctx context.Context, template *models.Template,
) (*models.Template, error) {
	log := zerolog.Ctx(ctx)

	if template.Name == "" {
		return nil, BadRequestError(fmt.Errorf("field 'name' cannot be empty"))
	}

	templatedb := models.TemplateDB{
		Template:        *template,
		InternalVersion: 1,
	}
	templatedb.Template.CreatedAt = strfmt.DateTime(time.Now())
	if err := templatedb.ToDB(); err != nil {
		if errors.Is(err, models.ErrBadBodyFormat) {
			return nil, BadRequestError(err)
		}

		return nil, err
	}

	if err := HandleTemplateActivity(ctx, h.db, &templatedb, log); err != nil {
		return nil, err
	}

	q := database.SQLInsertNoPKey(templatedb).
		Suffix("RETURNING " + strings.Join(models.TemplateColumns, ","))

	tx, err := database.Begin(ctx, h.db)
	if err != nil {
		return nil, err
	}
	defer tx.RollbackIfOpened(*log)

	if err := database.GetContext(ctx, tx, &templatedb, q, log); err != nil {
		if models.IsDuplicateError(err) {
			return nil, BadRequestError(fmt.Errorf("duplicate template name"))
		}

		return nil, err
	}
	if err := tx.Commit(); err != nil {
		return nil, err
	}

	return &templatedb.Template, nil
}

// Handle handles the incoming request.
func (h *AccountTemplateAddHandler) Handle(
	params template_management.AccountTemplateAddParams, principal *models.Principal,
) middleware.Responder {
	if principal.IsAdmin {
		return template_management.
			NewAccountTemplateAddUnauthorized().
			WithPayload(&models.Error{
				Message: "Admin cannot manage templates. Please use a non-admin user",
			})
	}
	if principal.Subject != params.Account {
		return template_management.
			NewAccountTemplateAddUnauthorized().
			WithPayload(&models.Error{
				Message: "User can manage its own templates only",
			})
	}
	params.Template.Account = params.Account
	ctx := params.HTTPRequest.Context()
	log := zerolog.Ctx(ctx)
	template, err := h.handle(params.HTTPRequest.Context(), params.Template)
	if err != nil {
		var httpError *HTTPErrorResponder
		if errors.As(err, &httpError) {
			return httpError
		}

		return InternalError(err, log, false)
	}

	return template_management.NewAccountTemplateAddCreated().WithPayload(template)
}
