package handlers

import (
	"context"
	"encoding/base64"
	"net/http"
	"testing"

	"github.com/Masterminds/squirrel"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/migration"
	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/testutils"
	"orus.io/orus-io/rednerd/utils"
)

func TestAccountTemplateUpdateHandler(t *testing.T) {
	db := database.GetTestDB(t, migration.Source)
	defer db.Close()

	handler := NewAccountTemplateUpdateHandler(db.DB)

	t.Run("Update non-existent template", func(t *testing.T) {
		log := testutils.GetLogger(t)
		ctx := log.WithContext(context.Background())
		tmpl, err := handler.handle(ctx, "janedoe", "hello", models.Template{})
		if assert.IsType(t, utils.HTTPError{}, err) {
			assert.Equal(t, http.StatusNotFound, err.(utils.HTTPError).StatusCode)
			assert.EqualError(t, err.(utils.HTTPError).Err, "No such template")
		}
		assert.Nil(t, tmpl)
	})

	t.Run("Update body", func(t *testing.T) {
		log := testutils.GetLogger(t)
		defer func() {
			_, _ = database.Exec(db, squirrel.Delete(models.AccountDBTable), &log)
			_, _ = database.Exec(db, squirrel.Delete(models.TemplateDBTable), &log)
		}()
		ctx := log.WithContext(context.Background())
		tdb := models.TemplateDB{
			Template: models.Template{
				Account: "janedoe", Name: "hello", Language: "mustache", Produces: "text/plain",
				BodyFormat: "text", Body: "Hello {{name}}",
			},
		}
		require.NoError(t, tdb.ToDB())

		_, err := database.Exec(db, database.SQLInsert(models.AccountDB{Name: "janedoe", IsUser: true}), &log)
		require.NoError(t, err)

		_, err = database.Exec(db, database.SQLInsertNoPKey(tdb), &log)
		require.NoError(t, err)

		tmpl, err := handler.handle(ctx, "janedoe", "hello", models.Template{
			Account: "janedoe", Name: "hello", Language: "mustache", Produces: "text/plain",
			BodyFormat: "text", Body: "Hello {{name}}!",
		})
		require.NoError(t, err)
		assert.Equal(t,
			base64.StdEncoding.EncodeToString([]byte("Hello {{name}}!")), tmpl.Body)
	})

	t.Run("Rename", func(t *testing.T) {
		log := testutils.GetLogger(t)
		defer func() {
			_, _ = database.Exec(db, squirrel.Delete(models.AccountDBTable), &log)
			_, _ = database.Exec(db, squirrel.Delete(models.TemplateDBTable), &log)
		}()
		ctx := log.WithContext(context.Background())
		tdb := models.TemplateDB{
			Template: models.Template{
				Account: "janedoe", Name: "hi", Language: "mustache", Produces: "text/plain",
				BodyFormat: "text", Body: "Hello {{name}}",
			},
		}
		require.NoError(t, tdb.ToDB())

		_, err := database.Exec(db, database.SQLInsert(models.AccountDB{Name: "janedoe", IsUser: true}), &log)
		require.NoError(t, err)

		_, err = database.Exec(db, database.SQLInsertNoPKey(tdb), &log)
		require.NoError(t, err)

		tmpl, err := handler.handle(ctx, "janedoe", "hi", models.Template{
			Account: "janedoe", Name: "hello", Language: "mustache", Produces: "text/plain",
			BodyFormat: "text", Body: "Hello {{name}}!",
		})
		require.NoError(t, err)
		require.Equal(t, "hello", tmpl.Name)

		require.NoError(t, database.Get(
			db,
			&tdb,
			squirrel.Select(models.TemplateDBNameColumn).From(models.TemplateDBTable),
			&log))
		assert.Equal(t, "hello", tdb.Name)
	})
	t.Run("Test Template update handler", func(t *testing.T) {
		h := NewAccountTemplateUpdateHandler(db.DB)
		expectedColumns := []string{"account", "data_schema", "description", "language", "locale", "metadata", "name", "produces", "static_metadata", "title", "version", "body"}
		require.Equal(t, expectedColumns, h.columns)
	})
}
