package handlers

import (
	"context"
	"strings"
	"time"

	"github.com/go-openapi/runtime/middleware"
	"github.com/go-openapi/strfmt"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/restapi/operations/user_management"
)

// NewUserAddHandler returns a UserAddHandler.
func NewUserAddHandler(db *sqlx.DB) *UserAddHandler {
	return &UserAddHandler{
		db: db,
	}
}

// UserAddHandler handles GET /user.
type UserAddHandler struct {
	db *sqlx.DB
}

func (h *UserAddHandler) handle(
	ctx context.Context, user *models.Userinfo,
) (*models.Userinfo, error) {
	var (
		log    = zerolog.Ctx(ctx)
		userdb = models.UserDB{
			Userinfo: *user,
		}
	)
	userdb.Userinfo.CreatedAt = strfmt.DateTime(time.Now().UTC())

	q := database.SQLInsert(userdb).
		Suffix("RETURNING " + strings.Join(models.UserinfoColumns, ","))

	tx, err := database.Begin(ctx, h.db)
	if err != nil {
		return nil, err
	}
	defer tx.RollbackIfOpened(*log)

	if _, err := database.ExecContext(
		ctx, tx,
		database.SQLInsert(models.AccountDB{Name: user.Username, IsUser: true}),
		log,
	); err != nil {
		return nil, err
	}
	if err := database.GetContext(ctx, tx, &userdb, q, log); err != nil {
		return nil, err
	}
	if err := tx.Commit(); err != nil {
		return nil, err
	}

	return &userdb.Userinfo, nil
}

// Handle handles the incoming request.
func (h *UserAddHandler) Handle(
	params user_management.UserAddParams, principal *models.Principal,
) middleware.Responder {
	if !principal.IsAdmin {
		return user_management.NewUserAddUnauthorized()
	}
	ctx := params.HTTPRequest.Context()
	log := zerolog.Ctx(ctx)
	user, err := h.handle(params.HTTPRequest.Context(), params.User)
	if err != nil {
		return InternalError(err, log, false)
	}

	return user_management.NewUserAddCreated().WithPayload(user)
}
